#ifndef __INTF_H__
#define __INTF_H__

#include <systemc.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <sys/time.h>
#include <unistd.h>

#include <string.h>
#include <stdio.h>

#include "net.h"

#define module intf

SC_MODULE( module) {
public:
	sc_in<sc_uint<32> > dr;
	sc_out<sc_uint<32> > dw;
	sc_out<sc_uint<32> > addr;
	sc_out<bool> cw, cr;
	sc_in<bool> clk;

	sc_uint<32> _dr;
	sc_uint<32> _dw;
	sc_uint<32> _addr;
	bool _cw, _cr;
	int cli_sock;			// -1 == not connected

	void work()
	{
		if (clk.posedge())
		{
			addr.write( _addr);
			dw.write( _dw);
			_dr = dr.read();
			if (_cw)
			{
				_cw = 0;
				cw.write( 1);
			}
			if (_cr)
			{
				_cr = 0;
				cr.write( 1);
			}
		}
		else
		{
			_dr = dr.read();
			cw.write( 0);
			if (cr.read())
			{
				int data, aa;

				data = _dr;
				aa = _addr;
				if (cli_sock != -1)
				{
					char buf[FRAME_LEN];
				
					sprintf( buf, "%x\n", data);
					write( cli_sock, buf, strlen( buf));
					printf( "wrote delayed data=%02X addr=%02X : [%s]\n", data, aa, buf);
				}
				else
				{
					printf( "data %02X at %02X lost because client disconnected\n", data, aa);
				}
				cr.write( 0);
			}
		}
	}
	
	void _write( int addr, int data)
	{
		_addr = addr;
		_dw = data;
		_cw = true;
	}
	
	int _read( int addr)
	{
		int result = 0;

		_addr = addr;
		_cr = true;
		result = _dr;

		return result;
	}

#define TIMEOUT 0

	void worker()
	{
		int s;
		struct sockaddr_in sa;
		socklen_t size;
		char buf[FRAME_LEN];
		int one = 1;
		int done = 0;

		s = socket( AF_INET, SOCK_STREAM, 0);
		printf( "got s=%d\n", s);
		setsockopt( s, SOL_SOCKET, SO_REUSEADDR, &one, sizeof( one));
		memset( &sa, 0, sizeof( sa));
		sa.sin_addr.s_addr = htonl( INADDR_ANY);
		sa.sin_port = htons( PORT);
		sa.sin_family = AF_INET;
		if (bind( s, (struct sockaddr *)&sa, sizeof( sa)) < 0)
		{
			printf( "bind error\n");
			goto end;
		}
		if (listen( s, 1) < 0)
		{
			printf( "listen error\n");
			goto end;
		}
		while (!done)
		{
			printf( "accepting on port %d\n", PORT);
			size = sizeof( sa);
			cli_sock = accept( s, (struct sockaddr *)&sa, &size);
			if (cli_sock < 0)
			{
				printf( "accept error\n");
				goto end;
			}
			printf( "accepted c=%d %s\n", cli_sock, inet_ntoa( sa.sin_addr));
			sprintf( buf, "hello %d - type help for help\n", cli_sock);
//			write( cli_sock, buf, strlen( buf));
			printf( "%s", buf);
		
			while (1)
			{
				ssize_t n, len;
				static int count = 0;
				int addr, data;
				fd_set rfds;
				int retval;
				struct timeval tv;
				int nsec = TIMEOUT, fd;
			
				FD_ZERO( &rfds);
				FD_SET( cli_sock, &rfds);
				tv.tv_sec = nsec;
				tv.tv_usec = 0;
				fd = cli_sock + 1;
				retval = select( fd, &rfds, NULL, NULL, &tv);
				if (retval == -1)
					perror( "select");
				else if (retval)
				{
					len = sizeof( buf);
					n = read( cli_sock, buf, len);
					if (n < len)
						buf[n] = 0;
					else
						buf[len - 1] = 0;
					printf( "recv returned %d : [%s]\n", (int)n, buf);
					if (n > 0)
					{
						if (!strncmp( buf, CMD_HELP, strlen( CMD_HELP)))
						{
							sprintf( buf, "%s\t\tterminate session\n", CMD_QUIT);
							write( cli_sock, buf, strlen( buf));
							printf( "%s", buf);
							sprintf( buf, "%s\t\tshow help\n", CMD_HELP);
							write( cli_sock, buf, strlen( buf));
							printf( "%s", buf);
							sprintf( buf, "%s\t\twrite data at addr\n", CMD_WRITE);
							write( cli_sock, buf, strlen( buf));
							printf( "%s", buf);
							sprintf( buf, "%s\t\tread data at addr\n", CMD_READ);
							write( cli_sock, buf, strlen( buf));
							printf( "%s", buf);
						}
						else if (!strncmp( buf, CMD_WRITE, strlen( CMD_WRITE)))
						{
							data = addr = 0;
							sscanf( buf, "%*s %x %x", &data, &addr);
							_write( addr, data);
							sprintf( buf, "wrote %02X at %02X\n", data, addr);
							printf( "%s", buf);
//							write( cli_sock, buf, strlen( buf));
						}
						else if (!strncmp( buf, CMD_READ, strlen( CMD_READ)))
						{
							addr = 0;
							sscanf( buf, "%*s %x", &addr);
							data = _read( addr);
							sprintf( buf, "read %02X at %02X -- delaying write\n", data, addr);
							printf( "%s", buf);
//							sprintf( buf, "%x\n", data);
//							write( cli_sock, buf, strlen( buf));
						}
						else if (!strncmp( buf, CMD_QUIT, strlen( CMD_QUIT)))
						{
							sprintf( buf, "good bye\n");
							printf( "%s", buf);
//							write( cli_sock, buf, strlen( buf));
							done = 1;
							break;
						}
						else
						{
							sprintf( buf, "unknown command (%d bytes) [%s]\n", (int)n, buf);
							printf( "%s", buf);
//							write( cli_sock, buf, strlen( buf));
						}
					}
					else if (n == 0)
					{
						printf( "end of file\n");
						break;
					}
					else
					{
						printf( "connection closed\n");
					//	done = 1;
						break;
					}
					count++;
				}
				else
				{
//					printf( "no data within %d seconds\n", nsec);
				}

				wait();
			}
			
			close( cli_sock);
			cli_sock = -1;
		}
		
		close( s);

end:		
		printf( "the end\n");
		sc_stop();
	}

	SC_CTOR( module) {
		cli_sock = -1;
		_dw = 0;
		_cw = 0;
		_cr = 0;
		_addr = 0;
		dw.initialize( 0);
		cw.initialize( 0);
		SC_METHOD( work);
		sensitive << clk;
		SC_CTHREAD( worker, clk.pos());
	}
};

#undef module

#endif/*__INTF_H__*/
